# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import logging
import sys
from typing import Iterable

import click
import nr.fs
from nr.stream import Stream
from termcolor import colored

from shut.changelog.manager import ChangelogManager
from shut.commands.commons.bump import make_bump_command, VersionBumpData, VersionRef
from shut.commands.pkg.bump import PackageBumpData
from shut.model import MonorepoModel, Project
from shut.model.version import get_commit_distance_version, parse_version, Version
from shut.utils.text import substitute_ranges
from . import mono
from .checks import check_monorepo
from .update import update_monorepo

logger = logging.getLogger(__name__)


class MonorepoBumpdata(VersionBumpData[MonorepoModel]):

  def update(self, new_version: Version) -> None:
    self.obj.version = new_version
    vfiles = update_monorepo(self.obj, dry=self.args.dry, indent=1)
    return vfiles.abspaths(self.obj.get_directory())

  def get_version_refs(self) -> Iterable[VersionRef]:
    yield from super().get_version_refs()
    if self.obj.release.single_version:
      for package in self.project.packages:
        yield from PackageBumpData(self.args, self.project, package).get_version_refs()

  def bump_to_version(self, target_version: Version) -> Iterable[str]:
    changed_files = list(super().bump_to_version(target_version))

    if not self.obj.release.single_version:
      return changed_files

    inter_deps = list(self.obj.get_inter_dependencies())
    if not inter_deps:
      return changed_files

    print()
    print(f'bumping {len(inter_deps)} mono repository inter-dependency(-ies)')
    for filename, refs in Stream.groupby(inter_deps, lambda d: d.filename, collect=list):
      print(f'  {colored(nr.fs.rel(filename), "cyan")}:')

      with open(filename) as fp:
        content = fp.read()

      for ref in refs:
        value = content[ref.version_start:ref.version_end]
        print(f'    {ref.package_name} {value} → ^{target_version}')

      content = substitute_ranges(
        content,
        ((ref.version_start, ref.version_end, f'^{target_version}') for ref in refs),
      )
      if not self.args.dry:
        with open(filename, 'w') as fp:
          fp.write(content)

      changed_files.append(filename)

    if self.args.tag and inter_deps and self.args.skip_update:
      logger.warning('bump requires an update in order to automatically tag')

    return changed_files

  def get_snapshot_version(self) -> Version:
    return get_commit_distance_version(
      self.obj.directory,
      self.obj.version,
      self.obj.get_tag(self.obj.version)) or self.obj.version

  def get_changelog_managers(self) -> Iterable[ChangelogManager]:
    yield from super().get_changelog_managers()
    if self.obj.release.single_version:
      for package in self.project.packages:
        yield from PackageBumpData(self.args, self.project, package).get_changelog_managers()


mono.command()(make_bump_command(MonorepoBumpdata, MonorepoModel))



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import shlex
import subprocess as sp
import sys
from typing import List, Optional, Set, Tuple

import click
from nr.stream import concat
from termcolor import colored

from shut.model.package import PackageModel
from shut.model.requirements import Requirement, RequirementsList, VendoredRequirement
from . import pkg
from .. import project


def collect_installable_requirements(package: bool, inter_deps: bool, extra: Set[str]) -> RequirementsList:
  reqs = RequirementsList()

  # Pip does not understand "test" as an extra and does not have an option to
  # install test requirements.
  if 'test' in extra:
    reqs += package.test_requirements

  reqs.append(VendoredRequirement(VendoredRequirement.Type.Path, package.get_directory()))
  reqs += package.requirements.vendored_reqs()

  if project.monorepo and inter_deps:
    # TODO(NiklasRosenstein): get_inter_dependencies_for() does not currently differentiate
    #   between normal, test and extra requirements.
    project_packages = {p.name: p for p in project.packages}
    for ref in project.monorepo.get_inter_dependencies_for(package):
      dep = project_packages[ref.package_name]
      if not ref.version_selector.matches(dep.version):
        print('note: skipping inter-dependency on package "{}" because the version selector '
              '{} does not match the present version {}'
              .format(dep.name, ref.version_selector, dep.version), file=sys.stderr)
      else:
        reqs.insert(0, VendoredRequirement(VendoredRequirement.Type.Path, dep.get_directory()))

  return reqs


def run_install(
  pip: Optional[str],
  reqs: List[Tuple[str, RequirementsList]],
  develop: bool,
  extra: Set[str],
  upgrade: bool,
  quiet: bool,
  dry: bool,
  pip_args: Optional[str],
) -> None:

  pip_bin = shlex.split(os.getenv('PIP', pip or 'python -m pip'))
  command = pip_bin + ['install']
  for directory, req in reqs:
    command += req.to_pip_args(directory, develop)
  if upgrade:
    command.append('--upgrade')
  if quiet:
    command.append('--quiet')
  command += shlex.split(pip_args) if pip_args else []

  if dry:
    print(' '.join(map(shlex.quote, command)))
  else:
    sys.exit(sp.call(command))


@pkg.command()
@click.option('--develop/--no-develop', default=True,
  help='Install in develop mode (default: true)')
@click.option('--inter-deps/--no-inter-deps', default=True,
  help='Install package inter dependencies from inside the same monorepo (default: true)')
@click.option('--extra', help='Specify one or more extras to install.')
@click.option('-U', '--upgrade', is_flag=True, help='Upgrade all packages (forwarded to pip install).')
@click.option('-q', '--quiet', is_flag=True, help='Quiet install')
@click.option('--pip', help='Override the command to run Pip. Defaults to "python -m pip" or the PIP variable.')
@click.option('--pip-args', help='Additional arguments to pass to Pip.')
@click.option('--dry', is_flag=True, help='Print the Pip command to stdout instead of running it.')
def install(develop, inter_deps, extra, upgrade, quiet, pip, pip_args, dry):
  """
  Install the package using `python -m pip`. If the package is part of a mono repository,
  inter-dependencies will be installed from the mono repsitory rather than from PyPI.

  The command used to invoke Pip can be overwritten using the `PIP` environment variable.
  """

  package = project.load_or_exit(expect=PackageModel)
  extra = set((extra or '').split(','))
  reqs = collect_installable_requirements(package, inter_deps, extra)
  run_install(pip, [(package.get_directory(), reqs)], develop, extra,upgrade, quiet, dry, pip_args)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import ast
import os
import re
import warnings
from typing import Dict, List, Optional

from databind.core import datamodel, field
from nr.stream import chain, concat

from shut.utils.ast import load_module_members
from .abstract import AbstractProjectModel
from .linter import LinterConfiguration
from .publish import PublishConfiguration
from .requirements import BaseRequirement, Requirement, RequirementsList
from .test import TestDriver
from .version import Version


class PackageError(Exception):
  pass


def _get_file_in_directory(directory: str, prefix: str, preferred: List[str]) -> Optional[str]:
  """
  Returns a file in *directory* that is either in the *preferred* list or starts with
  specified *prefix*.
  """

  choices = []
  for name in sorted(os.listdir(directory)):
    if name in preferred:
      break
    if name.startswith(prefix):
      choices.append(name)
  else:
    if choices:
      return choices[0]
    return None

  return os.path.join(directory, name)


@datamodel
class InstallConfiguration:
  @datamodel
  class InstallHooks:
    before_install: List[str] = field(altname='before-install', default_factory=list)
    after_install: List[str] = field(altname='after-install', default_factory=list)
    before_develop: List[str] = field(altname='before-develop', default_factory=list)
    after_develop: List[str] = field(altname='after-develop', default_factory=list)

    def any(self):
      return any((self.before_install, self.after_install, self.before_develop, self.after_develop))

  hooks: InstallHooks = field(default_factory=InstallHooks)


@datamodel
class PackageModel(AbstractProjectModel):
  modulename: Optional[str] = None
  description: Optional[str] = None
  readme: Optional[str] = None
  wheel: Optional[bool] = True
  universal: Optional[bool] = None
  typed: Optional[bool] = False
  requirements: RequirementsList = field(default_factory=RequirementsList)
  test_requirements: RequirementsList = field(altname='test-requirements', default_factory=RequirementsList)
  extra_requirements: Dict[str, RequirementsList] = field(altname='extra-requirements', default_factory=dict)
  source_directory: str = field(altname='source-directory', default='src')
  exclude: List[str] = field(default_factory=lambda: ['test', 'tests', 'docs'])
  entrypoints: Dict[str, List[str]] = field(default_factory=dict)
  classifiers: List[str] = field(default_factory=list)
  keywords: List[str] = field(default_factory=list)
  # TODO: Data files

  install: InstallConfiguration = field(default_factory=InstallConfiguration)
  linter: LinterConfiguration = field(default_factory=LinterConfiguration)
  publish: PublishConfiguration = field(default_factory=PublishConfiguration)
  test_driver: TestDriver = field(altname='test-driver', default=None)

  def get_modulename(self) -> str:
    if self.modulename:
      return self.modulename

    # Check for PEP-561 stub packages.
    # See also https://mypy.readthedocs.io/en/latest/installed_packages.html#making-pep-561-compatible-packages
    if self.name.endswith('-stubs'):
      return self.name[:-6].replace('-', '_') + '-stubs'

    return self.name.replace('-', '_')

  def get_python_requirement(self) -> Optional[Requirement]:
    return next(filter(lambda x: isinstance(x, Requirement) and x.package == 'python', self.requirements), None)

  def has_vendored_requirements(self) -> bool:
    """
    Returns #True if the package has any vendored requirements.
    """

    return any(chain(
      self.requirements.vendored_reqs(),
      self.test_requirements.vendored_reqs(),
      concat(l.vendored_reqs() for l in self.extra_requirements.values())
    ))

  def is_universal(self) -> bool:
    """
    Checks if the package is a universal Python package (i.e. it is Python 2 and 3 compatible)
    by testing the `$.requirements.python` version selector. If none is specified, the package
    is also considered universal.
    """

    if self.universal is not None:
      return self.universal

    python_requirement = self.get_python_requirement()
    if not python_requirement:
      return True

    # TODO (@NiklasRosenstein): This method of detecting if the version selector
    #   selects a Python 2 and 3 version is very suboptimal.
    has_2 = re.search(r'\b2\b|\b2\.\b', str(python_requirement))
    has_3 = re.search(r'\b3\b|\b3\.\b', str(python_requirement))
    return bool(has_2 and has_3)

  def get_python_package_metadata(self) -> 'PythonPackageMetadata':
    """
    Returns a #PythonPackageMetadata object for this #PackageModel. This object can be
    used to inspect the author and version information that is defined in the package
    source code.
    """

    return PythonPackageMetadata(
      os.path.join(os.path.dirname(self.filename), self.source_directory),
      self.get_modulename())

  def get_readme_file(self) -> Optional[str]:
    """
    Returns the absolute path to the README for this package.
    """

    directory = os.path.dirname(self.filename)

    if self.readme:
      return os.path.abspath(os.path.join(directory, self.readme))

    return _get_file_in_directory(
      directory=directory,
      prefix='README.',
      preferred=['README.md', 'README.rst', 'README.txt', 'README'])

  def get_license_file(self) -> Optional[str]:
    """
    Returns the absolute path to the LICENSE file for this package.
    """

    return _get_file_in_directory(
      directory=os.path.dirname(self.filename),
      prefix='LICENSE.',
      preferred=['LICENSE', 'LICENSE.txt', 'LICENSE.rst', 'LICENSE.md'])

  def get_py_typed_file(self) -> Optional[str]:
    if not self.typed:
      return None

    directory = self.get_python_package_metadata().package_directory
    return os.path.join(directory, 'py.typed')

  def get_publish_config(self) -> PublishConfiguration:
    if self.project and self.project.monorepo and self.project.monorepo.publish:
      return self.project.monorepo.publish
    return self.publish

  # AbstractProjectModel

  def get_name(self) -> str:
    return self.name

  def get_version(self) -> Optional[Version]:
    return self.version

  def get_tag(self, version: Version) -> str:
    tag_format = self.release.tag_format
    if self.project and self.project.monorepo and '{name}' not in tag_format:
      tag_format = '{name}@' + tag_format
    return tag_format.format(name=self.name, version=version)


class PythonPackageMetadata:
  """
  Represents the metadata of a Python package on disk.
  """

  def __init__(self, source_directory: str, modulename: str) -> None:
    self.source_directory = source_directory
    self.modulename = modulename
    self._filename = None
    self._author = None
    self._version = None

  @property
  def filename(self) -> str:
    """
    Returns the file that contains the package metadata in the Python source code. This is
    usually the module filename, the package `__init__.py` or `__version__.py`.
    """

    if self._filename:
      return self._filename

    parts = self.modulename.split('.')
    prefix = os.sep.join(parts[:-1])
    choices = [
      parts[-1] + '.py',
      os.path.join(parts[-1], '__version__.py'),
      os.path.join(parts[-1], '__init__.py'),
    ]
    for filename in choices:
      filename = os.path.join(self.source_directory, prefix, filename)
      if os.path.isfile(filename):
        self._filename = filename
        return filename

    raise PackageError('Entry file for package "{}" could not be determined'
                       .format(self.modulename))

  @property
  def package_directory(self) -> str:
    """
    Returns the Python package directory. Raises a #ValueError if this metadata represents
    just a single Python module.
    """

    if self.is_single_module:
      raise ValueError('this package is in module-only form')

    return os.path.dirname(self.filename) or '.'

  @property
  def is_single_module(self) -> bool:
    basename = os.path.basename(self.filename)
    return basename not in ('__init__.py', '__version__.py')

  @property
  def author(self) -> str:
    if not self._author:
      self._load_metadata()
    return self._author

  @property
  def version(self) -> str:
    if not self._version:
      self._load_metadata()
    return self._version

  def _load_metadata(self) -> None:
    members = load_module_members(self.filename)

    author = None
    version = None

    if '__version__' in members:
      try:
        version = ast.literal_eval(members['__version__'])
      except ValueError:
        version = '<Non-literal expression>'

    if '__author__' in members:
      try:
        author = ast.literal_eval(members['__author__'])
      except ValueError:
        author = '<Non-literal expression>'

    self._author = author
    self._version = version



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
from typing import Generic, Iterable, List, T, Type

from nr.stream import concat

from shut.builders import Builder
from shut.model import AbstractProjectModel
from shut.model.target import Target, TargetId
from shut.utils.type_registry import TypeRegistry

__all__ = [
  'Publisher',
  'PublisherProvider',
  'register_publisher_provider',
  'get_publishers',
]


class Publisher(Target, metaclass=abc.ABCMeta):

  @abc.abstractmethod
  def get_description(self) -> str:
    pass

  @abc.abstractmethod
  def get_build_dependencies(self) -> Iterable[TargetId]:
    """
    Return the IDs of build targets that this publisher depends on.
    """

  @abc.abstractmethod
  def publish(self, files: List[str], test: bool, verbose: bool) -> bool:
    """
    Run the publishing logic. The builders resolved from #get_build_dependencies() are
    passed to this function. They will already be built when this method is called.
    """


class PublisherProvider(Generic[T], metaclass=abc.ABCMeta):

  @abc.abstractmethod
  def get_publishers(self, obj: T) -> Iterable[Publisher]:
    """
    Return the publishers provided by this plugin.
    """


registry = TypeRegistry[PublisherProvider[AbstractProjectModel]]()


def register_publisher_provider(type_: Type[T], provider_class: Type[PublisherProvider[T]]) -> None:
  registry.put(type_, provider_class)


def get_publishers(obj: T) -> Iterable[Publisher]:
  return concat(provider().get_publishers(obj) for provider in registry.for_type(type(obj)))



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import sys
from typing import Any, Dict, List, TextIO, Tuple, Type, TypeVar, Union

import nr.fs
import yaml
from databind.core import ConversionError, datamodel, field, Registry
from databind.json import from_json, to_json, registry as json_registry
from nr.stream import Stream

registry = Registry(json_registry)
registry.set_option(datamodel, 'skip_defaults', True)
T = TypeVar('T')
ExcInfo = Tuple


def get_existing_file(directory: str, choices: List[str]) -> bool:
  for fn in choices:
    path = os.path.join(directory, fn)
    if os.path.isfile(path):
      return path
  return None


class Unexpected(Exception):

  def __init__(self, expected, got):
    self.expected = expected
    self.got = got

  def __str__(self):
    return f'expected: {self.expected}, got {self.got}'


class Project:
  """
  Loads package and mono repo configuration files and caches them to ensure that
  the same filename is never loaded into a different model object.
  """

  monorepo_filenames = ['monorepo.yml', 'monorepo.yaml']
  package_filenames = ['package.yml', 'package.yaml']

  Unexpected = Unexpected

  def __init__(self):
    self._cache: Dict[str, 'AbstractProjectModel'] = {}
    self.subject: 'AbstractProjectModel' = None
    self.monorepo: MonorepoModel = None
    self.packages: List[PackageModel] = []
    self.invalid_packages: List[Tuple[str, ExcInfo]] = []

  def load(
    self,
    directory: str = '.',
    expect: Type['AbstractProjectModel'] = None,
  ) -> 'AbstractProjectModel':
    """
    Loads all project information from *directory*. This searches in all parent directories
    for a package or monorepo configuration, then loads all resources that belong to the
    project.
    """

    directory = os.path.abspath(directory)
    monorepo_fn = None
    package_fn = None

    # TODO(NiklasRosenstein): Iterate parent dirs until match is found.
    for dirname in [directory]:
      package_fn = get_existing_file(dirname, self.package_filenames)
      if package_fn:
        break
      monorepo_fn = get_existing_file(dirname, self.monorepo_filenames)
      if monorepo_fn:
        break

    if package_fn:
      monorepo_fn = get_existing_file(os.path.dirname(os.path.dirname(package_fn)),
                                      self.monorepo_filenames)

    if monorepo_fn:
      self.subject = self._load_monorepo(monorepo_fn)
    if package_fn:
      self.subject = self._load_package(package_fn)

    if expect and not isinstance(self.subject, expect):
      raise Unexpected(expect, type(self.subject))

    return self.subject

  def load_or_exit(self, *args, **kwargs):
    try:
      return self.load(*args, **kwargs)
    except Unexpected as exc:
      if exc.expected == MonorepoModel:
        sys.exit('error: not in a mono repository context')
      elif exc.expected == PackageModel:
        sys.exit('error: not in a package context')
      else:
        raise

  def _load_object(self, filename: str, type_: Type[T]) -> T:
    filename = os.path.normpath(os.path.abspath(filename))
    if filename in self._cache:
      obj = self._cache[filename]
      assert isinstance(obj, type_), 'type mismatch: have {} but expected {}'.format(
        type(obj).__name__, type_.__name__)
      return obj
    with open(filename) as fp:
      data = yaml.safe_load(fp)
    #node_collector = NodeCollector()
    obj = self._cache[filename] = from_json(type_, data, registry=registry)
    obj.filename = filename
    obj.project = self
    #obj.unknown_keys = list(Stream.concat(
    #    (x.locator.append(k) for k in x.unknowns)
    #    for x in node_collector.nodes))
    return obj

  def _load_monorepo(self, filename: str) -> 'MonorepoModel':
    self.monorepo = self._load_object(filename, MonorepoModel)

    # Load packages in that monorepo.
    directory = os.path.dirname(filename)
    for item_name in os.listdir(directory):
      package_fn = get_existing_file(os.path.join(directory, item_name), self.package_filenames)
      if package_fn:
        try:
          self._load_package(package_fn)
        except ConversionError as exc:
          self.invalid_packages.append((item_name, sys.exc_info()))

    return self.monorepo

  def _load_package(self, filename: str) -> 'PackageModel':
    package = self._load_object(filename, PackageModel)
    if package not in self.packages:
      self.packages.append(package)
    return package


def dump(obj: Any, file_: Union[str, TextIO]) -> None:
  if isinstance(file_, str):
    with nr.fs.atomic_file(file_, 'w') as fp:
      dump(obj, fp)
  else:
    data = to_json(obj, registry=registry)
    yaml.safe_dump(data, file_, sort_keys=False)


from .abstract import AbstractProjectModel
from .monorepo import MonorepoModel
from .package import PackageModel



import click

from shut.commands.pkg.install import collect_installable_requirements, run_install
from shut.model.monorepo import MonorepoModel
from shut.model.requirements import RequirementsList
from . import mono, project


@mono.command()
@click.option('--develop/--no-develop', default=True,
  help='Install in develop mode (default: true)')
@click.option('--extra', help='Specify one or more extras to install.')
@click.option('-U', '--upgrade', is_flag=True, help='Upgrade all packages (forwarded to pip install).')
@click.option('-q', '--quiet', is_flag=True, help='Quiet install')
@click.option('--pip', help='Override the command to run Pip. Defaults to "python -m pip" or the PIP variable.')
@click.option('--pip-args', help='Additional arguments to pass to Pip.')
@click.option('--dry', is_flag=True, help='Print the Pip command to stdout instead of running it.')
def install(develop, extra, upgrade, quiet, pip, pip_args, dry):
  """
  Install all packages in the monorepo using`python -m pip`.

  The command used to invoke Pip can be overwritten using the `PIP` environment variable.
  """

  project.load_or_exit(expect=MonorepoModel)
  extra = set((extra or '').split(','))

  reqs = []
  for package in project.packages:
    reqs.append((package.get_directory(), collect_installable_requirements(package, False, extra)))

  run_install(pip, reqs, develop, extra,upgrade, quiet, dry, pip_args)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import re
import subprocess
import sys
from typing import List

import click
import nr.fs
import requests
from nr.utils.git import Git
from termcolor import colored

from shut.model import PackageModel
from shut.model.version import parse_version
from . import project, shut

logger = logging.getLogger(__name__)


def _call(*args, **kwargs):
  proc = subprocess.Popen(*args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **kwargs)
  stdout, _ = proc.communicate()
  if proc.returncode != 0:
    print(colored(stdout.decode(), 'red'))
    sys.exit(proc.returncode)


@shut.group()
def conda_forge():
  """
  Utility to update conda-forge recipes.
  """


@conda_forge.command()
@click.argument('package_name', required=False)
@click.argument('version', type=parse_version, required=False)
@click.option('-f', '--force', is_flag=True, help='force push to the Git repository')
def update_feedstock(package_name, version, force):
  """
  Update a conda-forge feedstock from a PyPI release. You can explicitly specify a *package_name*
  and *version*, or rely on the metadata from the package configuration file in your cwd.

  This command will clone the feedstack from the conda-forge organization on GitHub, then
  update thr version and sha256 reference and push a new branch to the repository.
  """

  if package_name or version:
    if not (package_name and version):
      sys.exit('error: need package_name AND version, OR neither')
    version = str(version)
  else:
    package = project.load_or_exit(expect=PackageModel)
    package_name, version = package.name, str(package.version)
    del package

  repo_name = 'conda-forge/{}-feedstock'.format(package_name)
  clone_url = 'git@github.com:' + repo_name
  branch_name = 'v' + version

  s_repo_name = colored(repo_name, 'cyan')
  s_package_name = colored(package_name, 'yellow')
  s_clone_url = colored(clone_url, 'cyan')
  s_branch_name = colored(branch_name, 'yellow')

  print(f'Fetching PyPI record for {s_package_name} ...')
  url = f'https://pypi.org/pypi/{package_name}/json'
  pypi_data = requests.get(url).json()
  if version not in pypi_data['releases']:
    sys.exit(f'error: no release on PyPI for {s_package_name} v{version}.')

  sdist = next((x for x in pypi_data['releases'][version] if x['packagetype'] == 'sdist'), None)
  if not sdist:
    logger.error('No sdist on PyPI for %s v%s.', s_package_name, version)
    sys.exit(1)
  sha256 = sdist['digests']['sha256']

  print(f'Updating {s_repo_name} from PyPI package {s_package_name} v{version}.')

  with nr.fs.tempdir() as tmpdir:

    print(f'Cloning {s_clone_url} ...')
    git = Git().clone(tmpdir.name, clone_url)
    git.create_branch(branch_name)

    recipe_path = nr.fs.join(tmpdir.name, 'recipe', 'meta.yaml')
    with open(recipe_path) as fp:
      recipe = fp.read()

    recipe = re.sub(r'(version = ")([^"]+)"', r'\g<1>' + version + '"', recipe)
    recipe = re.sub(r'(sha256:\s*")([^"]+)"', r'\g<1>' + sha256 + '"', recipe)

    with open(recipe_path, 'w') as fp:
      fp.write(recipe)

    print()
    subprocess.check_call(['git', '--no-pager', 'diff'], cwd=tmpdir.name)
    print()

    git.add(['.'])
    git.commit(f'Updating recipe for {package_name} v{version}.')

    print(f'pushing branch {s_branch_name} to {s_clone_url} ...')
    git.push('origin', branch_name, force=force)
